from . import hashing
from . import pickling
from . import preprocessing
from . import diskcache
from . import mongodbcache
from functools import update_wrapper


def persist(func=None,
            cache='file://persist/',
            funcname=None,
            key=None,
            storekey=False,
            pickle=pickling.pickle,
            unpickle=pickling.unpickle,
            hash=hashing.hash,
            unhash=None):
    """Function decorator for persistent memoisation

    Store the output of a function permanently, and use previously stored
    results instead of recomputing them.

    To use this, decorate the desired function with `@persist`.
    Or to customise the way this memoisation is done, decorate with
    `@persist(<args>)` and specify custom parameters.

    Parameters
    ----------
    cache : str, optional
        The address of the cache in which the outputs of this function should
        be stored.  If it starts with 'file://', then the remainder of the
        string should be the path to a location on the local file system in
        which the results will be stored; this may be a relative path.  If it
        starts with 'mongodb://' then the remainder of the string should be the
        URL of the pypersist MongoDB server in which the results will be stored.
        If it does not contain '://' then 'file://' will be added at the
        beginning.  Default is 'file://persist'.
    funcname : str, optional
        A string that uniquely describes this function.  If the same `cache` is
        used for several memoised functions, they should all have different
        `funcname` values.  Default is the name of the function.
    key : function(args -> object), optional
        Function that takes the arguments given to the memoised function, and
        returns a key that uniquely identifies those arguments.  Two sets of
        arguments should have the same key only if they produce the same output
        when passed into the memoised function.  Default returns a sorted tuple
        describing the arguments along with their names.
    storekey : bool, optional
        Whether to store the key along with the output when a result is stored.
        If True, the key will be checked when recalling a previously computed
        value, to check for hash collisions.  If False, two keys will produce
        the same output whenever their `hash` values are the same.  Default is
        False.
    pickle : function(object -> str), optional
        Function that converts the output of the function to a string for
        storage.  Should be the inverse of `unpickle`.  If `storekey` is true,
        this will also be used to store the key, and should do so without
        newline characters.  Default uses the `pickle` module and base 64
        encoding.
    unpickle : function(str -> object), optional
        Function that converts a string back to an object when retrieving a
        computed value from storage.  Should be the inverse of `pickle`.  If
        `storekey` is true, this will also be used to retrieve the key.
        Default uses the `pickle` module and base 64 encoding.
    hash : function(object -> str), optional
        Function that takes a key and produces a string that will be used to
        identify that key.  If this function is not injective, then `storekey`
        can be set to True to check for hash collisions.  The string should
        only contain characters safe for filenames.  Default uses SHA-256 and
        base 64 encoding, which has an extremely small chance of collision.
    unhash : function(str -> object), optional
        Function that, if specified, should be the inverse of `hash`.  If this
        is specified, it may be used whenever the keys of `cache` are
        requested.  Default is None.

    Attributes
    ----------
    cache : diskcache.Cache or mongodb.Cache
        Dictionary-like object that allows keys to be looked up and, if
        present, gives the previously computed value.  Values can be added and
        removed using the syntax `func.cache[key] = val` and
        `del func.cache[key]`.  If `storekey` is True or `unhash` is specified,
        this implements the collections.abc.MutableMapping abstract base class
        and we can iterate over its keys using `for key in func.cache`.

    Examples
    --------
    Simple persistence using default settings:

    >>> @persist
    ... def double(x):
    ...     return 2 * x
    >>> double(3)
    6
    >>> double(3)
    6
    >>> double.cache[(('x', 3),)]
    6

    Custom persistence using a simpler key, a descriptive filename, and writing
    human-readable files:

    >>> @persist(key=lambda x,y: (x,y),
    ...          hash=lambda k: '%s_to_the_power_of_%s' % k,
    ...          pickle=str,
    ...          unpickle=int)
    ... def power(x, y):
    ...     return x ** y
    >>> power(2,4)
    16
    >>> power(10,3)
    1000
    >>> power.cache[(2,4)]
    16

    """

    class persist_wrapper:

        def __init__(self, func):
            update_wrapper(self, func)
            self._func = func
            self._hash = hash
            self._unhash = unhash
            self._pickle = pickle
            self._unpickle = unpickle
            self._storekey = storekey
            if key is None:
                self._key = self.default_key
            else:
                self._key = key
            if funcname is None:
                self._funcname = func.__name__
            else:
                self._funcname = funcname

            # Determine which backend to use
            try:
                pos = cache.index('://')
                typestring = cache[:pos]
                path = cache[pos+len('://'):]
            except ValueError:
                # No backend specified: use disk
                typestring = 'file'
                path = cache

            cachetypes = {'file': diskcache,
                          'mongodb': mongodbcache}
            cachetype = cachetypes[typestring]

            if storekey or unhash:
                constr = cachetype.CacheWithKeys
            else:
                constr = cachetype.Cache

            self.cache = constr(self, path)

        def __call__(self, *args, **kwargs):
            key = self._key(*args, **kwargs)
            try:
                val = self.cache[key]
            except KeyError:
                val = self._func(*args, **kwargs)
                self.cache[key] = val
            return val

        def default_key(self, *args, **kwargs):
            return preprocessing.arg_tuple(self._func, *args, **kwargs)

        def clear(self):
            self.cache.clear()

    if func is None:
        # @persist(...)
        return persist_wrapper
    else:
        # @persist
        return persist_wrapper(func)



from .commoncache import HashCollisionError

from collections.abc import MutableMapping, Iterator
from os.path import join
import requests
import json


class Cache:
    """Dictionary-like object for saving function outputs to disk

    This cache, which can be used by the `persist` decorator in `persist.py`,
    stores computed values in a specified MongoDB database so that they can be
    restored later using a key.  Like a dictionary, a key-value pair can be
    added using `cache[key] = val`, looked up using `cache[key]`, and removed
    using `del cache[key]`.  The number of values stored can be found using
    `len(cache)`.

    A MongoDB cache might not store its keys, and therefore we cannot iterate
    through its keys as we can with a dictionary.  However, see
    `CacheWithKeys`.

    Parameters
    ----------
    func : persist_wrapper
        Memoised function whose results this is caching.  Options which are not
        specific to local disk storage, such as the key, hash, and pickle
        functions, are taken from this.
    url : str
        URL of the pypersist MongoDB database that will be used to store and load
        results.  The same database can be used for several different
        functions, since the function's `funcname` will be stored with each
        result.

    """

    def __init__(self, func, url):
        self._func = func

        # Use http if not specified
        if url.find('://') == -1:
            url = 'http://' + url
        self._url = join(url, self._func._funcname)
        self._headers = {'Content-type': 'application/json',
                         'Accept': 'text/plain'}

    def __getitem__(self, key):
        # Get hash and check it
        h = self._func._hash(key)
        if self._func._unhash:
            storedkey = self._func._unhash(h)
            if storedkey != key:
                raise HashCollisionError(storedkey, key)

        # Search for value in database
        db_item = self._get_db(h)
        if db_item:
            # Stored value found
            if self._func._storekey:
                # Check key
                keystring = db_item['key']
                storedkey = self._func._unpickle(keystring)
                if storedkey != key:
                    raise HashCollisionError(storedkey, key)
            # Use stored value
            val = self._func._unpickle(db_item['result'])
        else:
            # No value stored
            raise KeyError(key)

        return val

    def __setitem__(self, key, val):
        h = self._func._hash(key)
        new_item = {'funcname': self._func._funcname,
                    'hash': h,
                    'namespace': 'pypersist',
                    'result': self._func._pickle(val)}
        if self._func._storekey:
            new_item['key'] = self._func._pickle(key)
        r = requests.post(url=self._url,
                          headers=self._headers,
                          json=new_item)
        if r.status_code != 201:
            raise MongoDBError(r.status_code, r.reason)

    def __delitem__(self, key):
        # Get the item from the database
        h = self._func._hash(key)
        db_item = self._get_db(h)
        if db_item is None:
            raise KeyError(key)

        # Delete the item using its _id and _etag
        url = self._url + '/' + db_item['_id']
        headers = dict(self._headers)
        headers['If-Match'] = db_item['_etag']
        r = requests.delete(url=url, headers=headers)
        if r.status_code != 204:
            raise MongoDBError(r.status_code, r.reason)

    def __len__(self):
        db_items = self._get_db()
        if db_items:
            return db_items['_meta']['total']
        else:
            return 0

    def clear(self):
        """Delete all the results stored in this cache."""
        r = requests.delete(url=self._url)
        if r.status_code not in [204, 404]:
            raise MongoDBError(r.status_code, r.reason)

    def _get_db(self, hash=None):
        """Return all db items for this function, or one with this hash

        Queries the MongoDB database for entries with this function, and
        returns the resulting json data as a dictionary.  If a hash is
        specified, this will correspond to a single database item with entries
        '_id', '_etag', 'funcname', 'hash', 'result' and so on.  If no hash is
        specified, it will contain a list of all such items in the database in
        the '_items' entry, along with metadata in the '_meta' entry.

        If no appropriate item exists in the database, None is returned instead

        Parameters
        ----------
        hash : str, optional

        Returns
        -------
        dict or None

        """
        url = self._url
        if hash:
            url += '/' + hash
        r = requests.get(url=url)
        if r.status_code == 200:
            # Stored values found
            return json.loads(r.text)
        elif r.status_code == 404:
            # No value stored
            return None
        else:
            # Database error
            raise MongoDBError(r.status_code, r.reason)


class CacheWithKeys(Cache, MutableMapping):
    """Mutable mapping for saving function outputs to a MongoDB database

    This subclass of `Cache` can be used in place of `Cache` whenever
    `storekey` is True or `unhash` is defined, to implement the
    `MutableMapping` abstract base class.  This allows the cache to be used
    exactly like a dictionary, including the ability to iterate through all
    keys in the cache.

    """

    def __iter__(self):
        return self.KeysIter(self)

    class KeysIter(Iterator):
        """Iterator class for the keys of a `CacheWithKeys` object"""

        def __init__(self, cache):
            self._cache = cache
            assert(cache._func._storekey or cache._func._unhash)
            db_items = self._cache._get_db()
            if db_items:
                self._items = db_items['_items']
            else:
                self._items = []
            self._pos = 0

        def __next__(self):
            if self._pos >= len(self._items):
                raise StopIteration
            item = self._items[self._pos]
            self._pos += 1
            if self._cache._func._storekey:
                key = self._cache._func._unpickle(item['key'])
            else:
                assert(self._cache._func._unhash)
                key = self._cache._func._unhash(item['hash'])
            return key


class MongoDBError(Exception):
    """Exception for a problem interacting with a MongoDB database"""



from pickle import dumps
from hashlib import sha256
from base64 import urlsafe_b64encode
CHAR_ENCODING = 'utf-8'


def hash(key):
    """Return a string which is a hash of the argument given.

    It computes the SHA-256 sum of the key and returns it as a base 64 string.
    The string consists of alphanumeric characters, hyphens and underscores,
    and is precisely 43 characters long.

    Examples
    --------
    >>> hash('somestringkey123')
    'wXS1bv_UbdX4riiyyA3Djjo7JeiEfyGI7o1-hGMnkz0'
    >>> hash(3.141592654)
    'nAh_dG9CDZL7bAFWX7E3iUXN2HXZ5eUiYUzdCJXDH-k'
    >>> hash(None)
    'Tz_DSKgYlBpGTkFf_2udQWwd3DscZHQ4YdMo-8NFvNY'
    >>> key = (('arg1', [1,1,2,3,5,8,13]), ('x', 'hello'))
    >>> hash(key)
    '1TBQNjqeAKCcCBmy-Sk_T1Xm01juuHOWiKotF5WYeZ8'
    >>> hash('somestringkey123')
    'wXS1bv_UbdX4riiyyA3Djjo7JeiEfyGI7o1-hGMnkz0'

    """
    b = dumps(key)  # Pickle the key to a bytes object
    b = sha256(b).digest()  # Hash the bytes using sha-1
    b = urlsafe_b64encode(b)  # Convert the hash to a base64 string
    b = b[0:-1]  # Strip the final padding character ('=')
    s = b.decode(CHAR_ENCODING)  # Convert from bytes to string

    return s



from inspect import getfullargspec


def arg_tuple(func, *args, **kwargs):
    """Return a normalised tuple of arguments from args and kwargs

    This function checks that `func(*args, **kwargs)` is a valid function call,
    then converts all the non-keyword arguments to keyword arguments.  It then
    discards any supplied arguments that are equal to their default values
    (since they are unnecessary) and finally it sorts the arguments into
    alphabetical order by name.  The arguments are then retuend as a tuple of
    tuples, where each tuple is a pair containing the name of the argument
    followed by the value given.

    If `func` takes a variable number of arguments, any unnamed arguments will
    be included as a list with a name beginning with an asterisk.

    Parameters
    ----------
    func : function
        Function such that `func(*args, **kwargs)` is a valid call.
    args : list
        List of non-keyword arguments to be passed to `func`.
    kwargs : dict
        Dictionary of keyword arguments to be passed to `func`.

    Examples
    --------
    Tuple of arguments to built-in function `len`:

    >>> len('hello world')
    11
    >>> arg_tuple(len, 'hello world')
    (('obj', 'hello world'),)

    Tuple of arguments to user-defined function, with default argument being
    discarded:

    >>> def sum_of_three(x, a, m=2):
    ...     return x + a + m
    >>> sum_of_three(10, m=2, a=15)
    27
    >>> arg_tuple(sum_of_three, 10, m=2, a=15)
    (('a', 15), ('x', 10))

    """

    kwargs = kwargs.copy()

    spec = getfullargspec(func)

    # Check for too many arguments
    if len(args) > len(spec.args):
        if spec.varargs is None:
            func(*args, **kwargs)  # throws TypeError with useful message
        else:
            kwargs['*' + spec.varargs] = args[len(spec.args):]  # add varargs

    # Convert args to kwargs
    kwargs.update(dict(zip(spec.args, args)))

    # Remove any default arguments
    if spec.defaults is not None:
        for (arg, val) in zip(spec.args[-len(spec.defaults):], spec.defaults):
            if kwargs.get(arg) == val:
                kwargs.pop(arg)
    if spec.kwonlydefaults is not None:
        for arg in spec.kwonlydefaults:
            if kwargs.get(arg) == spec.kwonlydefaults[arg]:
                kwargs.pop(arg)

    # Return as a tuple
    out = tuple(sorted(kwargs.items()))
    return out



from pickle import dumps, loads
from base64 import urlsafe_b64encode, urlsafe_b64decode
CHAR_ENCODING = 'utf-8'


def pickle(obj):
    """Return a string representation of the object `obj`

    This function takes any object, and uses the `pickle` and `base64` modules
    to create a string which represents it.  This string consists only of
    alphanumeric characters, hyphens and underscores.  The object `obj` can
    later be reconstructed from this string using the `unpickle` function.

    Examples
    --------
    >>> pickle('Hello world')
    'gANYCwAAAEhlbGxvIHdvcmxkcQAu'
    >>> unpickle('gANYCwAAAEhlbGxvIHdvcmxkcQAu')
    'Hello world'

    """
    b = dumps(obj)  # object to bytes
    b64 = urlsafe_b64encode(b)  # bytes to base64 bytes
    s = b64.decode(CHAR_ENCODING)  # base64 bytes to string
    return s


def unpickle(string):
    """Restore an object from a string created by the `pickle` function

    If `string` was created by the `pickle` function in this file, then this
    function returns an object identical to the one that was used to create
    `string`.

    Examples
    --------
    >>> pickle('Hello world')
    'gANYCwAAAEhlbGxvIHdvcmxkcQAu'
    >>> unpickle('gANYCwAAAEhlbGxvIHdvcmxkcQAu')
    'Hello world'

    """
    b64 = string.encode(CHAR_ENCODING)  # string to base64 bytes
    b = urlsafe_b64decode(b64)  # base64 bytes to original bytes
    obj = loads(b)  # bytes to object
    return obj



from .commoncache import HashCollisionError

from collections.abc import MutableMapping, Iterator
from os import makedirs, remove, listdir
from os.path import exists, join


class Cache:
    """Dictionary-like object for saving function outputs to disk

    This cache, which can be used by the `persist` decorator in `persist.py`,
    stores computed values on disk in a specified directory so that they can be
    restored later using a key.  Like a dictionary, a key-value pair can be
    added using `cache[key] = val`, looked up using `cache[key]`, and removed
    using `del cache[key]`.  The number of values stored can be found using
    `len(cache)`.

    A disk cache might not store its keys, and therefore we cannot iterate
    through its keys as we can with a dictionary.  However, see
    `CacheWithKeys`.

    Parameters
    ----------
    func : persist_wrapper
        Memoised function whose results this is caching.  Options which are not
        specific to local disk storage, such as the key, hash, and pickle
        functions, are taken from this.
    dir : str
        Directory into which to save results.  The same directory can be used
        for several different functions, since a subdirectory will be created
        for each function based on its `funcname`.

    """

    def __init__(self, func, dir):
        self._func = func
        self._dir = join(dir, self._func._funcname)
        if not exists(self._dir):
            makedirs(self._dir)

    def __getitem__(self, key):
        fname = self._key_to_fname(key, OUT)
        if self._func._unhash:
            storedkey = self._fname_to_key(fname)
            if storedkey != key:
                raise HashCollisionError(storedkey, key)
        if exists(fname):
            if self._func._storekey:
                keyfname = self._key_to_fname(key, KEY)
                assert(exists(fname))
                keyfile = open(keyfname, 'r')
                keystring = keyfile.read()
                keyfile.close()
                storedkey = self._func._unpickle(keystring)
                if storedkey != key:
                    raise HashCollisionError(storedkey, key)
            file = open(fname, 'r')
            val = self._func._unpickle(file.read())
            file.close()
        else:
            raise KeyError(key)
        return val

    def __setitem__(self, key, val):
        if self._func._storekey:
            keyfname = self._key_to_fname(key, KEY)
            keyfile = open(keyfname, 'w')
            keyfile.write(self._func._pickle(key))
            keyfile.close()
        fname = self._key_to_fname(key, OUT)
        file = open(fname, 'w')
        file.write(self._func._pickle(val))
        file.close()

    def __delitem__(self, key):
        fname = self._key_to_fname(key, OUT)
        if exists(fname):
            remove(fname)
        else:
            raise KeyError(key)

    def __len__(self):
        # Number of files ending with '.out'
        return sum(fname.endswith(OUT) for fname in listdir(self._dir))

    def clear(self):
        """Delete all the results stored in this cache."""
        for f in listdir(self._dir):
            path = join(self._dir, f)
            # TODO: safety checks?
            remove(path)

    def _key_to_fname(self, key, ext):
        h = self._func._hash(key)
        return join(self._dir, h + ext)

    def _fname_to_key(self, fname):
        if fname.startswith(self._dir):
            fname = fname[len(self._dir + '/'):]  # remove directory
        h = fname[:fname.rfind('.')]  # remove extension
        return self._func._unhash(h)


class CacheWithKeys(Cache, MutableMapping):
    """Mutable mapping for saving function outputs to disk

    This subclass of `Cache` can be used in place of `Cache` whenever
    `storekey` is True or `unhash` is set, to implement the `MutableMapping`
    abstract base class.  This allows the cache to be used exactly like a
    dictionary, including the ability to iterate through all keys in the cache.

    """

    def __iter__(self):
        return self.KeysIter(self)

    class KeysIter(Iterator):
        """Iterator class for the keys of a `CacheWithKeys` object"""
        def __init__(self, cache):
            self._cache = cache
            self._pos = 0
            self._files = [fname
                           for fname in listdir(self._cache._dir)
                           if fname.endswith(OUT)]

        def __next__(self):
            if self._pos >= len(self._files):
                raise StopIteration
            fname = self._files[self._pos]
            if self._cache._func._unhash:
                # Unhash from filename
                key = self._cache._fname_to_key(fname)
            else:
                assert(self._cache._func._storekey)
                # Read key from file
                path = join(self._cache._dir, fname)
                path = path[:-len(OUT)] + KEY
                file = open(path, 'r')
                string = file.read()
                key = self._cache._func._unpickle(string)
                file.close()
            self._pos += 1
            return key


# Filename extensions
OUT = '.out'
KEY = '.key'
META = '.meta'



from setuptools import setup

setup(name='pypersist',
      version='0.2',
      description='Persistent memoisation framework for Python',
      url='https://github.com/mtorpey/pypersist',
      author='Michael Torpey',
      author_email='mct25@st-andrews.ac.uk',
      license='GPL',
      packages=['pypersist'],
      zip_safe=False)
